/*
 * pam_if.c
 *
 * Copyright (c) 1997, Pavel Kankovsky <kan@dcit.cz>, DCIT Ltd.
 * Based on the work of Elliot Lee, Marc Ewing, and Andrew G. Morgan.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 * 
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifdef linux
# define _SVID_SOURCE
# define _BSD_SOURCE
# define __USE_BSD
# define __USE_SVID
# define __USE_MISC
# define _GNU_SOURCE
# include <features.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>
#include <string.h>
#include <fnmatch.h>
#include <dlfcn.h>

#ifdef DEBUG
#include <assert.h>
#endif

/*
 * here, we make a definition for the externally accessible function
 * in this file (this definition is required for static a module
 * but strongly encouraged generally) it is used to instruct the
 * modules include file to define the function prototypes.
 */

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD

#include <security/pam_modules.h>
#include <security/_pam_macros.h>

/* some syslogging */

static void
_pam_log(int err, const char *format, ...)
{
  va_list args;

  va_start(args, format);
  openlog("PAM_if", LOG_CONS|LOG_PID, LOG_AUTH);
  vsyslog(err, format, args);
  va_end(args);
  closelog();
}

/* load module and call a function */

#define DEFAULT_MODULE_PATH "/lib/security/"

typedef int (*servicefn)(pam_handle_t *, int, int, const char **);

static int
relay_call(const char *mod_path, const char *proc,
           pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  char *mod_full_path = NULL;
  void *dlhandle = NULL;
  servicefn func = NULL;
  int rval;

  D(("called."));

  if (mod_path == NULL) {
    _pam_log(LOG_ERR, "null module path?!");
    return PAM_SERVICE_ERR;
  }
  
  if (mod_path[0] != '/') {
    mod_full_path = malloc(sizeof(DEFAULT_MODULE_PATH) + strlen(mod_path));
    sprintf(mod_full_path, DEFAULT_MODULE_PATH "%s", mod_path);
    mod_path = mod_full_path;
  }

  D(("module `%s'", mod_path));

  dlhandle = dlopen(mod_path, RTLD_NOW);
  if (dlhandle == NULL) {
    _pam_log(LOG_ERR, "unable to load %s: %s", mod_path, dlerror());
    return PAM_SERVICE_ERR;
  }

  D(("module added successfully"));

  _pam_drop(mod_full_path);

  func = (servicefn) dlsym(dlhandle, proc);
  if (func == NULL) {
    _pam_log(LOG_ERR, "unable to resolve symbol %s in %s", proc, mod_path);
    dlclose(dlhandle);
    return PAM_SERVICE_ERR;
  }

  D(("symbol %s resolved successfully", proc));

  rval = (*func)(pamh, flags, argc, argv);
  D(("(*func) result = %d", rval));

  dlclose(dlhandle);

  return rval;
}

/* check shell pattern */

static int
match(const char *pat, const char *str)
{
  if (!pat)
    return 1;
  if (!str)
    return 0;
  return !fnmatch(pat, str, 0);
}

/* relay function */

static int
relay(const char *proc, pam_handle_t *pamh, int flags,
      int argc, const char **argv)
{
  int rval, i;
  const char *a;
  int debug = 0;
  const char *userpat = NULL;
  const char *ttypat = NULL;
  const char *rhostpat = NULL;
  const char *val;

  D(("called."));

  for(i = 0; i < argc; ++i) {
    a = argv[i];
    if (!strcmp(a, "--"))
      { ++i; break; }
    else if (!strcmp(a, "debug"))
      debug = 1;
    else if (!strncmp(a, "user=", 5))
      userpat = a + 5;
    else if (!strncmp(a, "tty=", 4))
      ttypat = a + 4;
    else if (!strncmp(a, "rhost=", 5))
      rhostpat = a + 5;
    else {
      _pam_log(LOG_ERR, "unknown option: %s", a);
      return PAM_SERVICE_ERR;
    }
  }

  if (i >= argc) {
    _pam_log(LOG_ERR, "no module given", a);
    return PAM_SERVICE_ERR;
  }

  D(("userpat=%s, ttypat=%s, rhostpat=%s, i=%d",
    userpat, ttypat, rhostpat, i));

  rval = pam_get_user(pamh, &val, NULL);
  if (rval != PAM_SUCCESS) {
    _pam_log(LOG_ERR, "cannot retrieve username: %s",
             pam_strerror(pamh, rval));
    return PAM_SERVICE_ERR;
  }
  if (!match(userpat, val)) {
    D(("username mismatch"));
    if (debug)
      _pam_log(LOG_DEBUG, "username mismatch %s != %s",
               userpat, val);
    return PAM_IGNORE;
  }

  rval = pam_get_item(pamh, PAM_TTY, (const void **) &val);
  if (rval != PAM_SUCCESS) {
    _pam_log(LOG_ERR, "cannot retrieve tty: %s",
             pam_strerror(pamh, rval));
    return PAM_SERVICE_ERR;
  }
  if (!match(ttypat, val)) {
    D(("tty mismatch"));
    if (debug)
      _pam_log(LOG_DEBUG, "tty mismatch %s != %s",
               userpat, val);
    return PAM_IGNORE;
  }

  rval = pam_get_item(pamh, PAM_RHOST, (const void **) &val);
  if (rval != PAM_SUCCESS) {
    _pam_log(LOG_ERR, "cannot retrieve tty: %s",
             pam_strerror(pamh, rval));
    return PAM_SERVICE_ERR;
  }
  if (!match(rhostpat, val)) {
    D(("rhost mismatch"));
    if (debug)
      _pam_log(LOG_DEBUG, "rhost mismatch %s != %s",
               userpat, val);
    return PAM_IGNORE;
  }

  D(("match"));

  rval = relay_call(argv[i], proc, pamh, flags, argc-i-1, argv+i+1);
  D(("relay_call result=%d", rval));

  return rval;
}


/* define relay stubs */

#define RELAY_STUB(procname) \
\
PAM_EXTERN int \
procname(pam_handle_t *pamh, int flags, int argc, const char **argv) \
{ return relay(#procname, pamh, flags, argc, argv); }

RELAY_STUB(pam_sm_authenticate)
RELAY_STUB(pam_sm_setcred)
RELAY_STUB(pam_sm_acct_mgmt)
RELAY_STUB(pam_sm_open_session)
RELAY_STUB(pam_sm_close_session)
RELAY_STUB(pam_sm_chauthtok)


#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_listfile_modstruct = {
  "pam_if",
  pam_sm_authenticate,
  pam_sm_setcred,
  pam_sm_acct_mgmt,
  pam_sm_open_session,
  pam_sm_close_session,
  pam_sm_chauthtok,
};

#endif

/* end of module definition */

